/* (c) Copyright 1985 Cary Buchman and Daniel M. Berry */
/* (c) Copyright 1987 Berry Computer Scientists, Ltd. */
/************************************************************************* 
**************************************************************************
**                                                                      **
** this file includes routines that dump the current internal line      **
**                                                                      **
************************************************************************** 
*************************************************************************/ 

#include <stdio.h>
#include "TOKEN.h"
#include "lex.h"
#include "macros.h"

#define  max(a,b)       ((a<b) ? b : a)
#define  MAXZWC         100        /* maximum number pf respective zero
                                      width characters        */
#define  NFONT          255        /* defined also in width.c */
TOKENPTR new_token();              /* function return type    */

struct cnct
{
    char code[6];
    bool connect_before;
} connect[257] =
        {
#include "connect.h"
        };

extern   bool arabic_fonts[256];
extern   char *width_table[NFONT+1];
extern   char stretch_stage;
extern   int unit_width;

/************************************************************************* 
************************************************************************** 
**                                                                      **
** this routine dumps the current internal line while reversing         ** 
** the tokens of the specified lr direction                             **
** Was enhanced in order to deal with ZERO width characters, or         **
** ZERO horizontal movements (via the usage of '\z' for example).       **
** It was also optimized and made more efficient.                       **
**                                                                      **
** Johny                                                                **
************************************************************************** 
*************************************************************************/ 

dump_line(start,end,reverse_lr)

  TOKENPTR *start;
  TOKENPTR *end;
  bool     reverse_lr;

{

  bool     start_of_line = TRUE;

  TOKENPTR tmptr;
  TOKENPTR rev_start = NULL;      /* ptr to start of reversed token line */
  TOKENPTR rev_end   = NULL;      /* ptr to end of reversed token line   */
  TOKENPTR tokenptr  = *start;    /* next token to dump                  */
  TOKENPTR zerowc[MAXZWC];
  int      total_stretch = 0;
  int      i,j,hpos;
  
  while(tokenptr != NULL)
  {
    if  (tokenptr->lr != reverse_lr) 
    {                             /* if the language is written in it's
                                     natural direction                  */
      if (tokenptr->next == NULL)
        tokenptr->ending = NOT_END;
      put_token(tokenptr,start_of_line);
      start_of_line = FALSE;
      tokenptr = tokenptr->next;
    }
    else
    {
      while ((tokenptr != NULL) && (tokenptr->lr == reverse_lr))
      {
        if (tokenptr->ending) {            /* remark word endings */ 
            tokenptr->ending = NOT_END;
            tokenptr->begining = BEGINING;
        }
        if (tokenptr->begining) {
            tokenptr->ending = END;
            tokenptr->begining = NOT_BEGIN;
        }
        hpos = tokenptr->horizontal_pos;
        i = 0;
        while ((tokenptr->next != NULL) &&
               (tokenptr->next->horizontal_pos == hpos))
        {
            zerowc[i++] = tokenptr;
            tokenptr = tokenptr->next;
        }
        tmptr = tokenptr->next;
        if (i>0)        /* a ZERO-width character was detected */
        {
          int temp;
    
          zerowc[i] = tokenptr;
          for (j=i; j>=0; j--) {
            zerowc[j]->width = tokenptr->width;
            push_token(zerowc[j],&rev_start,&rev_end);
          }
          temp = zerowc[i]->ending;
          zerowc[i]->ending = zerowc[0]->ending;
          zerowc[0]->ending = NOT_END;
          zerowc[0]->begining = temp;
        }
        else push_token(tokenptr,&rev_start,&rev_end);
        tokenptr = tmptr;
      }

      recalculate_horizontal(rev_start,rev_end);

      if (tokenptr == NULL)            /* unmark word end at end of line */
        rev_end->ending = NOT_END;

      tmptr = rev_start;      /* dump reversed line */
      while (tmptr != NULL)
      {
        put_token(tmptr,start_of_line);
        start_of_line = FALSE;
        tmptr = tmptr->next;
      }

      free_line(&rev_start,&rev_end);
    }
  }

  free_line(start,end);
  
  return;
}

/*************************************************************************
**************************************************************************
**                                                                      **
** this routine reverses the specified internal line                    **
** It also recognizes ZERO width characters and preserves their         **
** order with the next letter.                                          **
**                                                                      **
** Johny                                                                **
**************************************************************************
*************************************************************************/


reverse_line(start,end,paper_width)

TOKENPTR *start;
TOKENPTR *end;
int      paper_width;

{
  int       line_length = (*end)->horizontal_pos + (*end)->width;
  int       new_indent  = paper_width - line_length - 1;
  int       adjustment  = new_indent - (*start)->horizontal_pos;
  TOKENPTR  tmptr = (*start);
  TOKENPTR  save_start;
  TOKENPTR  tmp_start   = NULL;
  TOKENPTR  tmp_end     = NULL;
  bool      begining;
  bool      ending;
  int       total_stretch;
  TOKENPTR  zerowc[MAXZWC];
  int       i,j,hpos;

  (*end)->ending = END;

  while ((tmptr = (*start)) != NULL)
  {
    *start = (*start)->next;

    begining = tmptr->begining;
    ending   = tmptr->ending;
    tmptr->begining = ending;
    tmptr->ending   = begining;
    hpos = tmptr->horizontal_pos;
    i = 0;
	/* enter the equaly hpos tokens into an array */
    while ((tmptr->next != NULL) && (tmptr->next->horizontal_pos == hpos))
    {
        zerowc[i++] = tmptr;
        tmptr = tmptr->next;
    }
    if (i>0) {
        int temp;

        zerowc[i] = tmptr;
        for (j=i-1; j>=0; j--)		/* update the start pointer */
            *start = (*start)->next;
        save_start = *start;
		/* update the width for each one of them */
        for (j=i; j>=0; j--) {
            zerowc[j]->width = tmptr->width;  /* take the last token width */
            zerowc[j]->horizontal_pos += adjustment;
            push_token(zerowc[j],&tmp_start,&tmp_end);
        }
        temp = zerowc[i]->ending;
        zerowc[i]->ending = zerowc[0]->ending;
        zerowc[0]->ending = NOT_END;
        zerowc[0]->begining = temp;
        *start = save_start;
    }
    else if (tmptr != NULL) {
        tmptr->horizontal_pos += adjustment;
        push_token(tmptr,&tmp_start,&tmp_end);
    }
  }

  *start = tmp_start;
  *end   = tmp_end;

  recalculate_horizontal(*start,*end);
  
  return;
}

/************************************************************************* 
************************************************************************** 
**                                                                      **
** this routine recalculates the horizontal motion in the specified     **
** internal line                                                        **
**                                                                      **
************************************************************************** 
*************************************************************************/ 

recalculate_horizontal(start,end)

  TOKENPTR  start;                  /* ptr to start of line */
  TOKENPTR  end  ;                  /* ptr to end of line   */

{

  TOKENPTR prev_token;
  TOKENPTR next_token;
  int      new_horizontal;
  int      prev_horizontal;
  int      start_horizontal;
  int      total_stretch;

  start_horizontal = start->horizontal_pos;
  prev_token = start; 
  prev_horizontal = end->horizontal_pos;
  new_horizontal  = prev_horizontal;
  
  if(prev_token->next != NULL)
  {
     next_token = prev_token->next;
     while (next_token != NULL)
     {
       new_horizontal = new_horizontal + (prev_token->horizontal_pos -
                        next_token->horizontal_pos) +
                        (prev_token->width) - (next_token->width);
       prev_token->horizontal_pos = prev_horizontal;
       prev_token = next_token;
       prev_horizontal = new_horizontal;
       next_token = prev_token->next;
     }
   }
   prev_token->horizontal_pos = new_horizontal;
 
   /* stretch Arabic words and reposition letters .. Johny */
   total_stretch = calculate_total_stretching(start,end);
   if (total_stretch < 0)       /* the line was shrinked!! */
       total_stretch = 0;
   stretch_letter(start,end,total_stretch);

   return;
}

/*************************************************************************
**************************************************************************
**                                                                      **
** This function should return the total stretching amount.             **
** It does that, by calculating the total amount of extra interword     **
** spaces.
**                                                                      **
** Johny                                                                **
**************************************************************************
*************************************************************************/

int
calculate_total_stretching (start,end)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */

{
  TOKENPTR prev_token;
  TOKENPTR next_token;
  TOKENPTR tokenptr  = start;
  int      total_stretch = 0;
  int      space_width;

  if ((tokenptr != NULL) && (tokenptr->next != NULL))
  {
      prev_token = tokenptr;
      next_token = tokenptr->next;
      while (next_token != NULL)
      {
        /* we have to calculate the space width for each word, as the
           point size can change at any point */
        space_width = (int) ((width_table[prev_token->font][0] * 
                      prev_token->point_size / unit_width) + 0.5);
        if (prev_token->ending) {
           total_stretch = total_stretch + (next_token->horizontal_pos -
                           prev_token->horizontal_pos) -
                           (prev_token->width + space_width);
        }
        prev_token = next_token;
        next_token = prev_token->next;
      }
      tokenptr = next_token;
  }

  return(total_stretch);
}

/*************************************************************************
**************************************************************************
**                                                                      **
** stretch the appropriate letter, based on the stretching stage.       **
** stretch_stage can be one of the following:							**
**		n : dont stretch at all.										**
**		f : stretch the final word only.								**
**		l : stretch the last word(s). the difference from the 'f'		**
**		    stage, is when the stretching amount is too big for one		**
**			word ... in that case, the remainder will be given to the	**
**			previous one.												**
**		a : stretch all of the Arabic words in the line, equally.		**
**																		**
** Johny.                                                               **
**                                                                      **
**************************************************************************
*************************************************************************/

stretch_letter (start,end,total_stretch)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */
int       total_stretch;

{
  int  filler_width;

  if (total_stretch == 0)
    return;
  switch (stretch_stage)
  {
  case 'n' :    /* NO stretching is done */
    {
      break;
    }
  case 'f' :    /* stretch the LAST connectable letter in the last word */
  case 'l' :
    {
      bool      finished = FALSE;
      TOKENPTR  tokenptr = start;
  
      while ((tokenptr != NULL) && !finished)
      {
        /* skip till the first arabic letter, or start of line */
        while ((tokenptr != NULL) &&
               (arabic_fonts[tokenptr->font] != ARABIC))
                      tokenptr = tokenptr->next;
        while ((tokenptr != NULL) && !finished &&
               (arabic_fonts[tokenptr->font] == ARABIC))
        {
          int  remainder = 0;
          char letter[4];
  
          sprintf(letter,"%c%c%c",tokenptr->char1,tokenptr->char2,
                                  tokenptr->char3);
          filler_width = width2("hy",tokenptr->point_size,tokenptr->font);
		  /* check if the letter is stretchable */
          if (connect[atoi(letter)].connect_before)
          {
            int  fillers_num = (int) (total_stretch / filler_width);
            int  limit = tokenptr->point_size;

            if ((stretch_stage == 'l') && (fillers_num > limit))
            {
                TOKENPTR tmptr = tokenptr;  /* save it. Usefull, in case
                                               we stretch the last (first)
                                               word in the line         */
                int      tmpstretch = total_stretch;

                tokenptr->fillers_num = limit;
                tokenptr->filler_width = filler_width;
                total_stretch = (tmpstretch - (limit*filler_width));
                /* skip to the next word, and execute another iteration */
                while ((tokenptr != NULL) && !(tokenptr->ending))
                    tokenptr = tokenptr->next;
                if (tokenptr == NULL)   /* this was the last word, so we
                                           can't do anything            */
                {
                    tmptr->fillers_num = fillers_num;
                    tmptr->filler_width = filler_width;
                    remainder = tmpstretch % filler_width;
                    recalculate_spaces(start,end,remainder);
                    finished = TRUE;
                    break;
                }
            }
            else {
                tokenptr->fillers_num = fillers_num;
                tokenptr->filler_width = filler_width;
                remainder = total_stretch % filler_width;
                recalculate_spaces(start,end,remainder);
                finished = TRUE;
                break;
            }
          }
          tokenptr = tokenptr->next;
        }
      }
      
      break;
    }
  case 'a' :    /* spread the stretching among the words */
    {
  
      int      ncw = 0;              /* number of connectable words     */
      int      spw = 0;              /* stretching per word             */
      int      last_word_remainder;  /* we stretch the last word more   */
      int      remainder = 0;
      bool     last_word = TRUE;     /* are we stretching the last word */
      TOKENPTR tokenptr = start;
  
      if ((ncw = number_connect_words(start,end)) == 0)
        break;                       /* there are NO stretchable words  */
      spw = (int) (total_stretch / ncw);
      while (tokenptr != NULL)
      {
        while ((tokenptr != NULL) &&
               (arabic_fonts[tokenptr->font] != ARABIC))
                      tokenptr = tokenptr->next;
        while ((tokenptr != NULL) &&
               (arabic_fonts[tokenptr->font] == ARABIC))
        {
          char letter[4];
  
          sprintf(letter,"%c%c%c",tokenptr->char1,tokenptr->char2,
                                  tokenptr->char3);
          if (connect[atoi(letter)].connect_before)
          {
            filler_width = width2("hy",tokenptr->point_size,tokenptr->font);
            if (last_word) {
              /* the total sum of remainders will be added to stretch
                 the last word */
              last_word_remainder = (total_stretch % ncw) +
                                    spw_remainders(start,end,spw) -
                                    (spw % filler_width);
              tokenptr->fillers_num = (int) ((spw+last_word_remainder) /
                                              filler_width);
              remainder = (spw + last_word_remainder) % filler_width;
              last_word = FALSE;
            }
            else tokenptr->fillers_num = (int) (spw / filler_width);
            tokenptr->filler_width = filler_width;
            while ((tokenptr != NULL) && !(tokenptr->ending))
              tokenptr = tokenptr->next;
          }
          if (tokenptr != NULL) 
              tokenptr = tokenptr->next;
        }
      }
      recalculate_spaces(start,end,remainder);
  
      break;
    }
  }

  return;

}


/*************************************************************************
**************************************************************************
**                                                                      **
** recalculates the inter words spaces, after stretching a specif-      **
** ied letter. It simply removes the  extra spaces added by  troff      **
** and spread the amount of remainder, which is the amount of poi-      **
** nts less that the filler width.                                      **
** initialization: hmove = 0                                            **
** when reaching an end of word: hmove = hmove - extra_space            **
** when reaching a letter with a filler: hmove = hmove + filler         **
**                                                                      **
** Johny.                                                               **
**                                                                      **
**************************************************************************
*************************************************************************/

recalculate_spaces (start,end,remainder)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */
int       remainder;

{
  TOKENPTR prev_token;
  TOKENPTR next_token;
  TOKENPTR tokenptr  = start;
  int      space_width=0, filler_width=0;
  int      extra_space=0, hmove=0;
  int      prev_horizontal;
  int      nw, rpw=0;                 /* remainder per word   */
  bool     last_word = TRUE;

  if ((nw = number_of_words(start,end)) == 0)
    return; 
  rpw = (int) (remainder / nw);

  if ((tokenptr != NULL) && (tokenptr->next != NULL))
  {
      prev_token = tokenptr;
      next_token = tokenptr->next;
      prev_horizontal = prev_token->horizontal_pos;
      while (next_token != NULL)
      {
        space_width = (int) ((width_table[prev_token->font][0] *
                            prev_token->point_size / unit_width) + 0.5);
        filler_width = width2("hy",next_token->point_size,next_token->font);
        if (prev_token->ending) {
           extra_space = (next_token->horizontal_pos -
                          prev_horizontal) -
                         (prev_token->width + space_width);
           hmove = hmove - extra_space + rpw;
           if (last_word) {
              hmove = hmove + (remainder % nw);
              last_word = FALSE;
           }
        }
        else if (prev_token->fillers_num > 0)
           hmove = hmove + prev_token->fillers_num * filler_width;
        prev_horizontal = next_token->horizontal_pos;
        next_token->horizontal_pos += hmove; 
        prev_token = next_token;
        next_token = next_token->next;
        if (next_token == NULL)
            prev_token->horizontal_pos += rpw;
      }
      tokenptr = next_token;
  }
}

/*************************************************************************
**************************************************************************
**                                                                      **
** return the number of words in the current line.                      **
** Johny.                                                               **
**                                                                      **
**************************************************************************
*************************************************************************/

int
number_of_words (start,end)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */
{
  TOKENPTR  tokenptr = start;
  int       nw = 0;               /* the number of Arabic words in line */

  while (tokenptr != NULL) {
    if (tokenptr->ending)
        nw++;
    tokenptr = tokenptr->next;
  }
  nw++;
  return(nw);
}

/*************************************************************************
**************************************************************************
**                                                                      **
** return the number of connectable words in the current line.          **
** a connectable word, is a word which contains a connectable           **
** letter.                                                              **
** Johny.                                                               **
**                                                                      **
**************************************************************************
*************************************************************************/

int
number_connect_words (start,end)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */
{
  TOKENPTR  tokenptr = start;
  int       ncw = 0;

        
  while (tokenptr != NULL) {
    bool    connectable = FALSE;
    char    letter[4];

    sprintf(letter,"%c%c%c",tokenptr->char1,tokenptr->char2,
                            tokenptr->char3);
    if ((arabic_fonts[tokenptr->font]) &&
        (connect[atoi(letter)].connect_before)) {
        connectable = TRUE;
        while ((tokenptr != NULL) && !(tokenptr->ending))
            tokenptr = tokenptr->next;
        if (tokenptr == NULL) {
            ncw++;
            break;
        }
    }
    if ((tokenptr->ending) && (arabic_fonts[tokenptr->font]) && connectable) {
        ncw++;
        connectable = FALSE;
    }
    tokenptr = tokenptr->next;
  }
  return(ncw);
}

/*************************************************************************
**************************************************************************
**                                                                      **
** returns the total remainders from all of the stretchable words.      **
** a word remainder is the word_stretchable_value modulu the filler     **
** width at that point.                                                 **
**                                                                      **
** Johny                                                                **
**************************************************************************
*************************************************************************/

int
spw_remainders (start,end,spw)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */
int       spw  ;
{
  TOKENPTR  tokenptr = start;
  int       spwr = 0;             /* stretch-per-word remainders */
  int       filler_width;

  while (tokenptr != NULL)
  {
    bool     connectable = FALSE;
    char     letter[4];
    TOKENPTR prev_token;

    sprintf(letter,"%c%c%c",tokenptr->char1,tokenptr->char2,
                            tokenptr->char3);
    if ((arabic_fonts[tokenptr->font]) &&
        (connect[atoi(letter)].connect_before))
    {
        connectable = TRUE;
        while ((tokenptr != NULL) && !(tokenptr->ending) &&
               (arabic_fonts[tokenptr->font]))
        {
            prev_token = tokenptr;
            tokenptr = tokenptr->next;
        }
        if (tokenptr == NULL) 
        {
            if (connectable) {
                filler_width = width2("hy",prev_token->point_size,
                                           prev_token->font);
                spwr += (spw % filler_width);
            }
            break; 
        }
    }
    if ((tokenptr->ending) && (arabic_fonts[tokenptr->font]) && connectable) {
        filler_width = width2("hy",tokenptr->point_size,tokenptr->font);
        spwr += (spw % filler_width);
        connectable = FALSE;
    }
    tokenptr = tokenptr->next;
  }
  return(spwr);
}


/* used for debugging ... Johny */
print_line(start,end)
TOKENPTR  *start;                 /* ptr to start of line */
TOKENPTR  *end  ;                 /* ptr to end of line   */
{

  TOKENPTR prev_token;
  TOKENPTR next_token;
  TOKENPTR tmptr = *start;
  
  while (tmptr != NULL) {
    printf("horiz:%d , vert:%d , fill_num:%d , char:%c%c%c , width:%d\n",
            tmptr->horizontal_pos,tmptr->vertical_pos,
            tmptr->fillers_num,tmptr->char1,
			tmptr->char2,tmptr->char3,tmptr->width);
    tmptr = tmptr->next;
  }
  printf("\n");
}

