/* (c) Copyright 1985 Cary Buchman and Daniel M. Berry */
/* (c) Copyright 1987 Berry Computer Scientists, Ltd. */
/*********************************************************
**********************************************************
**                                                      **
** this file contains a number of routines that work on **
** internal tokens and token lines                      **
**                                                      **
**********************************************************
*********************************************************/

#include  <stdio.h>
#include  "TOKEN.h"
#include  "TABLE.h"
#include  "macros.h"
#include  "lex.h"



extern int  out_font;                  /* current output font */
extern int  out_size;                  /* current output size */
extern int  out_horizontal;            /* current output horizontal position */
extern int  out_vertical;              /* current output vertical position */
extern char out_font_name[];           /* current output font name */
extern TABLENTRY out_fontable[];       /* currnet output font table */
extern int  in_font;                   /* current input font */
extern int  in_size;                   /* current input size */
extern int  in_horizontal;             /* current input horizontal position */
extern int  in_vertical;               /* current input vertical position */
extern char in_font_name[];            /* current input font name */
extern TABLENTRY in_fontable[];        /* currnet input font table */
extern bool in_lr;                     /* current input font direction */
extern bool direction_table[];         /* direction of font i */

int i;

/****************************************************************
*****************************************************************
**                                                             **
** this routine allocates and initializes a new internal token **
**                                                             **
*****************************************************************
****************************************************************/

TOKENPTR
new_token (toktyp,ptsize,font,horizontal,vertical,fontdir,beg_wrd,end_wrd,char1,
           char2,char3,f_name,width)

  int          toktyp;         /* token type                 */
  int          ptsize;         /* point size                 */
  int          font;           /* font number                */
  int          horizontal;     /* horizontal position        */
  int          vertical;       /* vertical position          */
  bool         fontdir;        /* font direction             */
  bool         beg_wrd;        /* begining of word indicator */
  bool         end_wrd;        /* end of word indicator      */
  char         char1;          /* 1st char of (abs) character token*/
  char         char2;          /* 2nd char of (abs) character token*/
  char         char3;          /* 3nd char of abs character token*/
  char         *f_name;        /* font name                  */
  int          width;          /* character width            */

{

  char  *calloc();
  TOKENPTR tokenptr = (TOKENTYPE *) calloc(1,sizeof(TOKENTYPE));

  if (tokenptr != NULL)
  {
    tokenptr->token_type    = toktyp;
    tokenptr->point_size    = ptsize;
    tokenptr->font          = font;
    tokenptr->horizontal_pos= horizontal;
    tokenptr->vertical_pos  = vertical;
    tokenptr->lr            = fontdir;
    tokenptr->begining      = beg_wrd;
    tokenptr->ending        = end_wrd;
	tokenptr->fillers_num	= NOFILLERS;	/* initialization ... Johny */
    tokenptr->char1         = char1;
    tokenptr->char2         = char2;
    tokenptr->char3         = char3;
    strcpy(tokenptr->font_name,f_name);
    tokenptr->width         = width;
    tokenptr->next          = NULL;
  }
  else
  {
    out_of_memory();
  }
 return (tokenptr);
}


/************************************************************************/
/************************************************************************/


/*
    this routine adds a token to the end of a line 
                                                     */

add_token (tokenptr,start,end)

  TOKENPTR     tokenptr;           /* token to be added        */
  TOKENPTR     *start;             /* ptr to start of line ptr */
  TOKENPTR     *end;               /* ptr to end of line ptr   */

{
  if (*start == NULL)
    *start = tokenptr;
  else {
    (*end)->next = tokenptr;
	tokenptr->next = NULL;        /* better to make sure it's NULL */
  }
  *end = tokenptr;
  return;
}

/************************************************************************/
/************************************************************************/

/*
    this routine pushes a token onto the front of a line 
                                                           */
push_token (tokenptr,start,end)

  TOKENPTR     tokenptr;            /* token to be push         */
  TOKENPTR     *start;              /* ptr to start of line ptr */
  TOKENPTR     *end;                /* ptr to end of line ptr   */

{

  if (*end == NULL) {
									/* reset the prev and next pointers ...
									   it's not sure that we call it ONLY
									   after new_token.  Johny  */
	(tokenptr)->next = NULL;
    *end = tokenptr;
  }
  else tokenptr->next = (*start);
  *start = tokenptr;
  return;
}

/*******************************************
********************************************
**                                        **
** this routine outputs an internal token ** 
**                                        **
********************************************
*******************************************/

put_token (tokenptr,start_of_line)

  TOKENPTR tokenptr;           /* token to output */
  bool     start_of_line;      /* indicates first character in line */

{
  if (start_of_line)
  {
    printf("H%d\n",tokenptr->horizontal_pos);
  }
 
  if (tokenptr->point_size != out_size)
  {
    printf ("s%d\n",tokenptr->point_size);
    out_size = tokenptr->point_size;
  }
  if (strcmp(tokenptr->font_name,out_font_name))
  {
    if (strcmp(tokenptr->font_name,out_fontable[tokenptr->font].name))
    {
      printf("x font %d %s\n",tokenptr->font,tokenptr->font_name);
      new_font(tokenptr->font,tokenptr->font_name,tokenptr->lr,
               out_fontable);
    }
    printf ("f%d\n",tokenptr->font);
    out_font = tokenptr->font;
    strcpy(out_font_name,tokenptr->font_name);
  } 

  if (tokenptr->vertical_pos != out_vertical)
  {
    printf ("V%d\n",tokenptr->vertical_pos);
    out_vertical = tokenptr->vertical_pos;
  }

  if (!start_of_line)
  {
    if (tokenptr->horizontal_pos < out_horizontal)
    {
      printf("H%d\n",tokenptr->horizontal_pos);
    }
    else if (tokenptr->horizontal_pos > out_horizontal) {
      int j =(tokenptr->horizontal_pos - out_horizontal);
      if (j < 100 && j > 9 && tokenptr->token_type == c_token){
              printf( "%d%c", j, tokenptr->char1);
              goto compress;
          }
      printf("h%d",(tokenptr->horizontal_pos - out_horizontal));
    }
  }


  switch (tokenptr->token_type) {
    case c_token: printf("c%c\n",tokenptr->char1); break;
    case C_token: printf("C%c%c\n",tokenptr->char1,tokenptr->char2); break;
    case N_token: printf("N%c%c%c\n",tokenptr->char1,tokenptr->char2,
						tokenptr->char3); break;
  }
compress: /* thanks to Mulli Bahr hnncx compressed to nnx */
  out_horizontal = tokenptr->horizontal_pos;

  /* added to insert the fillers ... Johny */
  if (tokenptr->fillers_num > 0) {
	int i;
	printf("h%d",tokenptr->width);
	printf("Chy\n");
	for (i=0; i<(tokenptr->fillers_num)-1; i++)
		printf("h%dChy\n",tokenptr->filler_width);
	printf("h%d",tokenptr->filler_width);
	out_horizontal += tokenptr->fillers_num * tokenptr->filler_width +
					  tokenptr->width;
  }

  if (tokenptr->ending == END)
    printf("w");

  return;
}


/*************************************************************
**************************************************************
**                                                          **
** this routine outputs a new page token and its associated **
** font and point size tokens and resets its font and size  **
** to null and 0 so as to force output of f and s commands  **
**                                                          **
**************************************************************
*************************************************************/

put_page_token(page_token)

  char  *page_token;          /* lex page token value */

{

  extern char  yytext[];     /* lex token value */
  
  printf("V%d\n",in_vertical); 
				/* ditroff guarantees and thus drivers
				require that before page_token, a V0 token! 
				Mulli Bahr, Oct, 86*/
  out_vertical = in_vertical;
  printf("%s\n",page_token);

  out_size = 0;                            /* reset out_size to zero ... */
  out_font_name[0] = '\0';                 /* reset out_font_name to empty
         string  so that the next token print will be forced to dump out s and f
         commands. On some devices the current size and page are not remembered
         across page boundaries, especially if there's a bunch of "x font XX"s.
         Besides the ditroff output always seems to have the s and f issued just
         before the text on each page */

  return;
}


/***********************************************************************/
/***********************************************************************/


/*
    this routine deallocates a line of tokens
                                                */

free_line(start,end)

  TOKENPTR  *start;           /* ptr to start of line */
  TOKENPTR  *end;             /* ptr to end of line   */

{
  TOKENPTR  tokenptr;         /* next element in line */

  while( (tokenptr = *start) != NULL)
  {
    *start = tokenptr->next;
    cfree(tokenptr);
  }
  *end = NULL;
  return;
}

/************************************************************************/ 
/************************************************************************/ 
